import asyncio
import logging
import websockets
from gabriel_protocol import gabriel_pb2
from collections import namedtuple


URI_FORMAT = 'ws://{host}:{port}'


logger = logging.getLogger(__name__)
websockets_logger = logging.getLogger(websockets.__name__)

# The entire payload will be printed if this is allowed to be DEBUG
websockets_logger.setLevel(logging.INFO)


ProducerWrapper = namedtuple('ProducerWrapper', ['producer', 'source_name'])


class WebsocketClient:
    def __init__(self, host, port, producer_wrappers, consumer):
        '''
        producer should take no arguments and return an instance of
        gabriel_pb2.InputFrame.
        consumer should take one gabriel_pb2.ResultWrapper and does not need to
        return anything.
        '''

        self._welcome_event = asyncio.Event()
        self._sources = {}
        self._running = True
        self._uri = URI_FORMAT.format(host=host, port=port)
        self.producer_wrappers = producer_wrappers
        self.consumer = consumer

    def launch(self):
        event_loop = asyncio.get_event_loop()

        # TODO remove this line once we stop supporting Python 3.5
        asyncio.set_event_loop(event_loop)

        try:
            self._websocket = event_loop.run_until_complete(
                websockets.connect(self._uri))
        except ConnectionRefusedError:
            logger.error('Could not connect to server')
            return

        consumer_task = asyncio.ensure_future(self._consumer_handler())
        tasks = [
            asyncio.ensure_future(self._producer_handler(
                producer_wrapper.producer, producer_wrapper.source_name))
            for producer_wrapper in self.producer_wrappers
        ]
        tasks.append(consumer_task)

        _, pending = event_loop.run_until_complete(asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED))
        for task in pending:
            task.cancel()
        logger.info('Disconnected From Server')

    def get_source_names(self):
        return self._sources.keys()

    def stop(self):
        self._running = False
        logger.info('stopping server')

    async def _consumer_handler(self):
        while self._running:
            try:
                raw_input = await self._websocket.recv()
            except websockets.exceptions.ConnectionClosed:
                return  # stop the handler
            logger.debug('Recieved input from server')

            to_client = gabriel_pb2.ToClient()
            to_client.ParseFromString(raw_input)

            if to_client.HasField('welcome'):
                self._process_welcome(to_client.welcome)
            elif to_client.HasField('response'):
                self._process_response(to_client.response)
            else:
                raise Exception('Empty to_client message')

    def _process_welcome(self, welcome):
        for source_name in welcome.sources_consumed:
            self._sources[source_name] = _Source(welcome.num_tokens_per_source)
        self._welcome_event.set()

    def _process_response(self, response):
        result_wrapper = response.result_wrapper
        if (result_wrapper.status == gabriel_pb2.ResultWrapper.SUCCESS):
            self.consumer(result_wrapper)
        elif (result_wrapper.status ==
              gabriel_pb2.ResultWrapper.NO_ENGINE_FOR_SOURCE):
            raise Exception('No engine for source')
        else:
            status = result_wrapper.Status.Name(result_wrapper.status)
            logger.error('Output status was: %s', status)

        if response.return_token:
            self._sources[response.source_name].return_token()

    async def _producer_handler(self, producer, source_name):
        '''
        Loop waiting until there is a token available. Then calls supplier to
        get the partially built FromClient to send.
        '''

        await self._welcome_event.wait()
        while self._running:
            assert source_name in self._sources, (
                'No engines consume source: {}'.format(source_name))
            source = self._sources[source_name]
            await source.get_token()

            input_frame = await producer()
            if input_frame is None:
                source.return_token()
                logger.info('Received None from producer')
                continue

            from_client = gabriel_pb2.FromClient()
            from_client.frame_id = source.get_frame_id()
            from_client.source_name = source_name
            from_client.input_frame.CopyFrom(input_frame)
            await self._send_from_client(from_client)
            logger.debug('num_tokens for %s is now %d', source_name,
                         source.get_num_tokens())
            source.next_frame()

    async def _send_from_client(self, from_client):
        try:
            await self._websocket.send(from_client.SerializeToString())
        except websockets.exceptions.ConnectionClosed:
            return  # stop the handler


class _Source:
    def __init__(self, num_tokens):
        self._num_tokens = num_tokens
        self._event = asyncio.Event()
        self._frame_id = 0

    def return_token(self):
        self._num_tokens += 1
        self._event.set()

    async def get_token(self):
        while self._num_tokens < 1:
            logger.debug('Waiting for token')
            self._event.clear()  # Clear because we definitely want wait
            await self._event.wait()

        self._num_tokens -= 1

    def get_num_tokens(self):
        return self._num_tokens

    def get_frame_id(self):
        return self._frame_id

    def next_frame(self):
        self._frame_id += 1



from gabriel_client.websocket_client import WebsocketClient
from gabriel_client.websocket_client import ProducerWrapper
import time
import logging


logger = logging.getLogger(__name__)


class MeasurementClient(WebsocketClient):
    def __init__(self, host, port, producer_wrappers, consumer, output_freq=10):
        super().__init__(host, port, producer_wrappers, consumer)

        self._source_measurements = {}
        self._output_freq = output_freq

    def _process_welcome(self, welcome):
        super()._process_welcome(welcome)
        start_time = time.time()
        for source_name in welcome.sources_consumed:
            source_measurement = _SourceMeasurement(
                start_time, self._output_freq)
            self._source_measurements[source_name] = source_measurement

    def _process_response(self, response):
        response_time = time.time()
        super()._process_response(response)
        if response.return_token:
            source_measurement = self._source_measurements[response.source_name]
            source_measurement.process_response(
                response.frame_id, response.source_name, response_time)

    async def _send_from_client(self, from_client):
        await super()._send_from_client(from_client)
        send_time = time.time()
        source_measurement = self._source_measurements[from_client.source_name]
        source_measurement.log_send(from_client.frame_id, send_time)

class _SourceMeasurement:
    def __init__(self, start_time, output_freq):
        self._count = 0
        self._send_timestamps = {}
        self._recv_timestamps = {}
        self._start_time = start_time
        self._interval_start_time = start_time
        self._output_freq = output_freq

    def process_response(self, frame_id, source_name, response_time):
        self._recv_timestamps[frame_id] = response_time
        self._count += 1

        if (self._count % self._output_freq) == 0:
            self._compute_and_print(source_name, response_time)
            self._interval_start_time = time.time()

    def _compute_and_print(self, source_name, response_time):
        print('Measurements for source:', source_name)
        overall_fps = _compute_fps(self._count, response_time, self._start_time)
        print('Overall FPS:', overall_fps)
        interval_fps = _compute_fps(
            self._output_freq, response_time, self._interval_start_time)
        print('Interval FPS:', interval_fps)

        total_rtt = 0
        for frame_id, received in self._recv_timestamps.items():
            sent = self._send_timestamps[frame_id]
            total_rtt += (received - sent)
            del self._send_timestamps[frame_id]

        print('Average RTT for interval:', total_rtt / self._output_freq)
        self._recv_timestamps.clear()

    def log_send(self, frame_id, send_time):
        self._send_timestamps[frame_id] = send_time


def _compute_fps(num_frames, current_time, start_time):
    return num_frames / (current_time - start_time)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gabriel-client",
    version="2.0.1",
    author="Roger Iyengar",
    author_email="ri@rogeriyengar.com",
    description="Client for Wearable Cognitive Assistance Applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://gabriel.cs.cmu.edu",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    license="Apache",
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
    install_requires=[
        "gabriel-protocol==2.0.1",
        "websockets==8.0; python_version>=\"3.6\"",
        "websockets==7.0; python_version<\"3.6\"",
        "opencv-python>=3, <5"
    ],
)



import asyncio
import os
from gabriel_protocol import gabriel_pb2
from gabriel_client.websocket_client import ProducerWrapper


_NUM_BYTES_FOR_SIZE = 4
_BYTEORDER = 'big'


def consumer(_):
    pass


class Source:
    def __init__(self, source_name):
        self._source_name = source_name
        self._frame_available = asyncio.Event()
        self._latest_input_frame = None

        self._constructor_pid = os.getpid()
        self._read, self._write = os.pipe()
        self._started_receiver = False

    def get_producer_wrapper(self):
        async def receiver():
            stream_reader = asyncio.StreamReader()
            def protocol_factory():
                return asyncio.StreamReaderProtocol(stream_reader)
            transport = await asyncio.get_event_loop().connect_read_pipe(
                protocol_factory, os.fdopen(self._read, mode='r'))

            # TODO do we need to close transport when filter gets garbage
            # collected?

            while True:
                size_bytes = await stream_reader.readexactly(
                    _NUM_BYTES_FOR_SIZE)
                size_of_message = int.from_bytes(size_bytes, _BYTEORDER)

                input_frame = gabriel_pb2.InputFrame()
                input_frame.ParseFromString(
                    await stream_reader.readexactly(size_of_message))
                self._latest_input_frame = input_frame
                self._frame_available.set()

        async def producer():
            if not self._started_receiver:
                self._started_receiver = True
                assert os.getpid() == self._constructor_pid
                asyncio.ensure_future(receiver())

            await self._frame_available.wait()

            # Clear because we are sending self._latest_input_frame
            self._frame_available.clear()

            return self._latest_input_frame

        return ProducerWrapper(producer=producer, source_name=self._source_name)

    def send(self, input_frame):
        serialized_message = input_frame.SerializeToString()

        size_of_message = len(serialized_message)
        size_bytes = size_of_message.to_bytes(_NUM_BYTES_FOR_SIZE, _BYTEORDER)

        num_bytes_written = os.write(self._write, size_bytes)
        assert num_bytes_written == _NUM_BYTES_FOR_SIZE, 'Write incomplete'

        num_bytes_written = os.write(self._write, serialized_message)
        assert num_bytes_written == size_of_message, 'Write incomplete'
